@@ -8,15 +8,13 @@
         char *member;
 #define VECTOR(member) \
         ASS_Vector member;
-#define BITMAPHASHKEY(member) \
-        BitmapHashKey member;
 #define END(typedefnamename) \
     } typedefnamename;
 
 #elif defined(CREATE_COMPARISON_FUNCTIONS)
 #undef CREATE_COMPARISON_FUNCTIONS
 #define START(funcname, structname) \
-    static unsigned funcname##_compare(void *key1, void *key2, size_t key_size) \
+    static bool funcname##_compare(void *key1, void *key2) \
     { \
         struct structname *a = key1; \
         struct structname *b = key2; \
@@ -27,28 +25,21 @@
             strcmp(a->member, b->member) == 0 &&
 #define VECTOR(member) \
             a->member.x == b->member.x && a->member.y == b->member.y &&
-#define BITMAPHASHKEY(member) \
-            bitmap_compare(&a->member, &b->member, sizeof(a->member)) &&
 #define END(typedefname) \
-            1; \
+            true; \
     }
 
 #elif defined(CREATE_HASH_FUNCTIONS)
 #undef CREATE_HASH_FUNCTIONS
 #define START(funcname, structname) \
-    static unsigned funcname##_hash(void *buf, size_t len) \
+    static uint32_t funcname##_hash(void *buf, uint32_t hval) \
     { \
-        struct structname *p = buf; \
-        unsigned hval = FNV1_32A_INIT;
+        struct structname *p = buf;
 #define GENERIC(type, member) \
         hval = fnv_32a_buf(&p->member, sizeof(p->member), hval);
 #define STRING(member) \
         hval = fnv_32a_str(p->member, hval);
 #define VECTOR(member) GENERIC(, member.x); GENERIC(, member.y);
-#define BITMAPHASHKEY(member) { \
-        unsigned temp = bitmap_hash(&p->member, sizeof(p->member)); \
-        hval = fnv_32a_buf(&temp, sizeof(temp), hval); \
-        }
 #define END(typedefname) \
         return hval; \
     }
@@ -60,25 +51,21 @@
 
 
 // describes an outline bitmap
-START(outline_bitmap, outline_bitmap_hash_key)
+START(bitmap, bitmap_hash_key)
     GENERIC(OutlineHashValue *, outline)
-    GENERIC(int, frx) // signed 10.22
-    GENERIC(int, fry) // signed 10.22
-    GENERIC(int, frz) // signed 10.22
-    GENERIC(int, fax) // signed 16.16
-    GENERIC(int, fay) // signed 16.16
-    // shift vector that was added to glyph before applying rotation
-    // = 0, if frx = fry = frx = 0
-    // = (glyph base point) - (rotation origin), otherwise
-    GENERIC(int, shift_x)
-    GENERIC(int, shift_y)
-    VECTOR(advance) // subpixel shift vector
-END(OutlineBitmapHashKey)
+    // quantized transform matrix
+    VECTOR(offset)
+    VECTOR(matrix_x)
+    VECTOR(matrix_y)
+    VECTOR(matrix_z)
+END(BitmapHashKey)
 
-// describe a clip mask bitmap
-START(clip_bitmap, clip_bitmap_hash_key)
-    STRING(text)
-END(ClipMaskHashKey)
+START(glyph_metrics, glyph_metrics_hash_key)
+    GENERIC(ASS_Font *, font)
+    GENERIC(double, size)
+    GENERIC(int, face_index)
+    GENERIC(int, glyph_index)
+END(GlyphMetricsHashKey)
 
 // describes an outline glyph
 START(glyph, glyph_hash_key)
@@ -88,47 +75,42 @@ START(glyph, glyph_hash_key)
     GENERIC(int, glyph_index)
     GENERIC(int, bold)
     GENERIC(int, italic)
-    GENERIC(unsigned, scale_x) // 16.16
-    GENERIC(unsigned, scale_y) // 16.16
-    VECTOR(outline) // border width, 26.6
-    GENERIC(unsigned, flags)    // glyph decoration flags
-    GENERIC(unsigned, border_style)
-    GENERIC(int, hspacing) // 16.16
+    GENERIC(unsigned, flags) // glyph decoration flags
 END(GlyphHashKey)
 
-START(glyph_metrics, glyph_metrics_hash_key)
-    GENERIC(ASS_Font *, font)
-    GENERIC(double, size)
-    GENERIC(int, face_index)
-    GENERIC(int, glyph_index)
-    GENERIC(unsigned, scale_x)
-    GENERIC(unsigned, scale_y)
-END(GlyphMetricsHashKey)
-
 // describes an outline drawing
 START(drawing, drawing_hash_key)
-    GENERIC(unsigned, scale_x)
-    GENERIC(unsigned, scale_y)
-    GENERIC(int, pbo)
-    VECTOR(outline)
-    GENERIC(unsigned, border_style)
-    GENERIC(int, hspacing)
-    GENERIC(int, scale)
-    GENERIC(unsigned, hash)
     STRING(text)
 END(DrawingHashKey)
 
+// describes an offset outline
+START(border, border_hash_key)
+    GENERIC(OutlineHashValue *, outline)
+    // outline is scaled by 2^scale_ord_x|y before stroking
+    // to keep stoker error in allowable range
+    GENERIC(int, scale_ord_x)
+    GENERIC(int, scale_ord_y)
+    VECTOR(border)  // border size in STROKER_ACCURACY units
+END(BorderHashKey)
+
 // describes post-combining effects
 START(filter, filter_desc)
     GENERIC(int, flags)
     GENERIC(int, be)
-    GENERIC(double, blur)
+    GENERIC(int, blur)
     VECTOR(shadow)
 END(FilterDesc)
 
+// describes glyph bitmap reference
+START(bitmap_ref, bitmap_ref_key)
+    GENERIC(Bitmap *, bm)
+    GENERIC(Bitmap *, bm_o)
+    VECTOR(pos)
+    VECTOR(pos_o)
+END(BitmapRef)
+
 #undef START
 #undef GENERIC
 #undef STRING
 #undef VECTOR
-#undef BITMAPHASHKEY
 #undef END
